// SPDX-License-Identifier: LGPL-3.0-or-later
// See Notices.txt for copyright information
use crate::{intern::Intern, prelude::*};

#[derive(Copy, Clone, PartialEq, Eq, PartialOrd, Ord, Hash, Debug)]
pub enum FormalKind {
    Assert,
    Assume,
    Cover,
}

impl FormalKind {
    pub fn as_str(self) -> &'static str {
        match self {
            Self::Assert => "assert",
            Self::Assume => "assume",
            Self::Cover => "cover",
        }
    }
}

#[track_caller]
pub fn formal_stmt_with_enable_and_loc(
    kind: FormalKind,
    clk: Expr<Clock>,
    pred: Expr<Bool>,
    en: Expr<Bool>,
    text: &str,
    source_location: SourceLocation,
) {
    crate::module::add_stmt_formal(crate::module::StmtFormal {
        kind,
        clk,
        pred,
        en,
        text: text.intern(),
        source_location,
    });
}

#[track_caller]
pub fn formal_stmt_with_enable(
    kind: FormalKind,
    clk: Expr<Clock>,
    pred: Expr<Bool>,
    en: Expr<Bool>,
    text: &str,
) {
    formal_stmt_with_enable_and_loc(kind, clk, pred, en, text, SourceLocation::caller());
}

#[track_caller]
pub fn formal_stmt_with_loc(
    kind: FormalKind,
    clk: Expr<Clock>,
    pred: Expr<Bool>,
    text: &str,
    source_location: SourceLocation,
) {
    formal_stmt_with_enable_and_loc(kind, clk, pred, true.to_expr(), text, source_location);
}

#[track_caller]
pub fn formal_stmt(kind: FormalKind, clk: Expr<Clock>, pred: Expr<Bool>, text: &str) {
    formal_stmt_with_loc(kind, clk, pred, text, SourceLocation::caller());
}

macro_rules! make_formal {
    ($kind:ident, $formal_stmt_with_enable_and_loc:ident, $formal_stmt_with_enable:ident, $formal_stmt_with_loc:ident, $formal_stmt:ident) => {
        #[track_caller]
        pub fn $formal_stmt_with_enable_and_loc(
            clk: Expr<Clock>,
            pred: Expr<Bool>,
            en: Expr<Bool>,
            text: &str,
            source_location: SourceLocation,
        ) {
            formal_stmt_with_enable_and_loc(
                FormalKind::$kind,
                clk,
                pred,
                en,
                text,
                source_location,
            );
        }
        #[track_caller]
        pub fn $formal_stmt_with_enable(
            clk: Expr<Clock>,
            pred: Expr<Bool>,
            en: Expr<Bool>,
            text: &str,
        ) {
            formal_stmt_with_enable(FormalKind::$kind, clk, pred, en, text);
        }
        #[track_caller]
        pub fn $formal_stmt_with_loc(
            clk: Expr<Clock>,
            pred: Expr<Bool>,
            text: &str,
            source_location: SourceLocation,
        ) {
            formal_stmt_with_loc(FormalKind::$kind, clk, pred, text, source_location);
        }
        #[track_caller]
        pub fn $formal_stmt(clk: Expr<Clock>, pred: Expr<Bool>, text: &str) {
            formal_stmt(FormalKind::$kind, clk, pred, text);
        }
    };
}

make_formal!(
    Assert,
    hdl_assert_with_enable_and_loc,
    hdl_assert_with_enable,
    hdl_assert_with_loc,
    hdl_assert
);

make_formal!(
    Assume,
    hdl_assume_with_enable_and_loc,
    hdl_assume_with_enable,
    hdl_assume_with_loc,
    hdl_assume
);

make_formal!(
    Cover,
    hdl_cover_with_enable_and_loc,
    hdl_cover_with_enable,
    hdl_cover_with_loc,
    hdl_cover
);

pub trait MakeFormalExpr: Type {}

impl<T: Type> MakeFormalExpr for T {}

#[hdl]
pub fn formal_global_clock() -> Expr<Clock> {
    #[hdl_module(extern)]
    fn formal_global_clock() {
        #[hdl]
        let clk: Clock = m.output();
        m.annotate_module(BlackBoxInlineAnnotation {
            path: "fayalite_formal_global_clock.v".intern(),
            text: r"module __fayalite_formal_global_clock(output clk);
    (* gclk *)
    reg clk;
endmodule
"
            .intern(),
        });
        m.verilog_name("__fayalite_formal_global_clock");
    }
    #[hdl]
    let formal_global_clock = instance(formal_global_clock());
    formal_global_clock.clk
}

#[hdl]
pub fn formal_reset() -> Expr<AsyncReset> {
    #[hdl_module(extern)]
    fn formal_reset() {
        #[hdl]
        let rst: AsyncReset = m.output();
        m.annotate_module(BlackBoxInlineAnnotation {
            path: "fayalite_formal_reset.v".intern(),
            text: r"module __fayalite_formal_reset(output rst);
    reg rst;
    (* gclk *)
    reg gclk;
    initial rst = 1;
    always @(posedge gclk)
        rst <= 0;
endmodule
"
            .intern(),
        });
        m.verilog_name("__fayalite_formal_reset");
    }
    #[hdl]
    let formal_reset = instance(formal_reset());
    formal_reset.rst
}
